unit virtualRenderer;

interface

  uses {$IFDEF WIN32}windows,graphics{$ENDIF}, mathObjects,sysutils;
type
  {$IFNDEF WIN32}
    TPoint= record
              x,y:integer;
            end;
    TRect = record
              case Integer of
                0: (Left, Top, Right, Bottom: Integer);
                1: (TopLeft, BottomRight: TPoint);
            end;
  {$ENDIF}
    TPoint64F= record
              x,y:double;
            end;

  TMoveTo=procedure (x,y:integer) of object;
  TLineTo=TMoveTo;
  TTextOut=procedure (X, Y: Integer; const Text: string) of object;
  TTextWidth=function (const Text: string): Integer of object;
  TTextHeight=TTextWidth;
  TSetColor=procedure (color:TColor) of object;

  TOutputFunctions = record
    moveTo: TMoveTo;
    lineTo:TLineTo;
    textout: TTextOut;
    textWidth: TTextWidth;
    textHeight: TTextHeight;
    setColor:TSetColor;
  end;
  TOutputOptions = record
    showCoordinationPos,showGrid: boolean;
    showBorder: boolean;
    showExtrema: boolean;
    showArrows: boolean;
    coordStrikeLength: longint;
    intervalSizeX,intervalSizeY: longint;
    null:TPoint;
    size:TRect;
    factorX,factorY:double;
    gridColor,coordColor, borderColor, extremaColor: tcolor;
  end;
  TVirtualRenderer=class
  protected
    currentColor: TColor;
  public
    outputFunctions: TOutputFunctions;
    outputOptions: TOutputOptions;

    mathSystem:TMathSystem;
    constructor create;
    destructor destroy;override;
//    procedure setShowCoordinationPos(const v: boolean);

    procedure renderCoordinates;
    procedure render(term: pterm); overload;
    procedure render(term: pterm; color: tcolor); overload;
    procedure renderWithVar(term: pterm; changeVar: PVariable); overload;

    function pixelPosToCoordPos(const x,y:longint):tpoint64f;overload;
    function pixelPosToCoordPos(const p:tpoint):tpoint64f;overload;
    function coordPosToPixelPos(const x,y:double):tpoint;overload;
    function coordPosToPixelPos(const p:tpoint64f):tpoint;overload;

    procedure centerTo(const x,y:double); overload;
    procedure centerTo(const p:TPoint64F); overload;
    procedure zoomTo(const x,y:double; const f: double); overload;
    procedure zoomTo(const p:TPoint64F; const f: double); overload;
  end;
const Epsilon=1e-30;
implementation
constructor TVirtualRenderer.create;
begin
  with outputOptions do begin
    factorX:=50;
    factorY:=-50;
    showCoordinationPos:=true;
    showExtrema:=true;
    intervalSizeX:=50;
    intervalSizeY:=50;
    coordStrikeLength:=10;
    extremaColor:=clLime;

    showArrows:=true;
  end;

  outputOptions.gridColor:=clGray;
  outputOptions.coordColor:=clBlack;
  outputOptions.borderColor:=clBlack;

  outputFunctions.textOut:=nil;
  outputFunctions.textWidth:=nil;
  outputFunctions.textHeight:=nil;
  outputFunctions.setColor:=nil;

  currentColor:=clBlack;
end;
{procedure TVirtualRenderer.setShowCoordinationPos(const v: boolean);
begin
  showCoordinationPos:=v;
  renderCoordinates;
end; }
procedure TVirtualRenderer.renderCoordinates;
var //i:integer;
    x,y:longint;
    n: string;
begin
  with outputOptions, outputFunctions do begin
    if showBorder then begin
      if assigned(setcolor) then
        setColor(borderColor);
      moveTo(size.Left,size.top);
      lineTo(size.Right,size.top);
      lineTo(size.Right,size.Bottom);
      lineTo(size.Right,size.Bottom);
      lineTo(size.Left,size.Bottom);
      lineTo(size.Left,size.top);
    end;

    if assigned(setcolor) then
      setColor(coordColor);

    moveto(size.left,null.y);
    lineto(size.right,null.y);
    moveto(null.x,size.top);
    lineto(null.x,size.bottom);


    x:=size.left - size.left mod intervalSizeX + null.x mod intervalSizeX;
    if x<size.left then inc(x,intervalSizeX);
    while (x<size.right) do begin
  //    n:=format('%.3g',[(x-null.x)/factorX]); //floatToStr((x-null.x)/factorX);
      n:=floatToStr((x-null.x)/factorX);
      if n<>'0' then begin
        if showGrid and assigned(setColor) then begin
          setColor(gridColor);
          moveTo(x,size.top);
          lineTo(x,size.bottom);
          setColor(coordColor);
        end;
        if showCoordinationPos and assigned(textOut) and assigned(textWidth) then
          textout(x-textwidth(n) div 2,null.y+coordStrikeLength div 2+2,n);
      end;
      moveTo(x,null.y-coordStrikeLength div 2);
      lineTo(x,null.y+coordStrikeLength div 2+1);
      inc(x,intervalSizeX);
    end;

    y:=size.top - size.top mod intervalSizeY + null.y mod intervalSizeY;
    if y<size.top then inc(y,intervalSizeY);
    while (y<size.bottom) do begin
      n:=floatToStr((y-null.y)/factorY);
      if n<>'0' then begin
        if showGrid and assigned(setColor) then begin
          setColor(gridColor);
          moveto(size.left,y);
          lineto(size.right,y);
          setColor(coordColor);
        end;
        if showCoordinationPos and assigned(textOut) and assigned(textHeight) then
          textout(null.x+coordStrikeLength div 2+2,y-textHeight(n) div 2,n);
      end;
      moveTo(null.x-coordStrikeLength div 2,y);
      lineTo(null.x+coordStrikeLength div 2+1,y);

      inc(y,intervalSizeY);
    end;

    if showArrows then begin
//      moveTo(null.x-coordStrikeLength div 2,size.top+1+coordStrikeLength);
      if null.y>size.top+coordStrikeLength then begin
        moveTo(null.x-coordStrikeLength div 2,size.top+coordStrikeLength);
        lineTo(null.x,size.top+1);
        lineTo(null.x+coordStrikeLength div 2,size.top+1+coordStrikeLength);
      end;
      if null.x<size.right-coordStrikeLength then begin
        moveTo(size.right-coordStrikeLength,null.y-coordStrikeLength div 2);
        lineTo(size.right-1,null.y);
        lineTo(size.right-1-coordStrikeLength,null.y+coordStrikeLength div 2);
      end;
    end;
  end;
end;

procedure TVirtualRenderer.render(term: pterm);
begin
  render(term,clBlack);
end;
procedure TVirtualRenderer.render(term: pterm;color: tcolor);
var func: PFunction;
    selectedValues: array of integer;
    changeVars: TVariableArray;
    i,j:integer;
    r,g,b,fac:double;
    br,bg,bb: integer;

begin
  if assigned(outputFunctions.setcolor) then begin
    currentColor:=color;
    outputFunctions.setcolor(color);
  end;

  if term^.typ<>  tt2Function then begin
    renderWithVar(term,mathSystem.findVariable('x'));
    exit;
  end;

  func:=PFunction(term);
  if func^.paramVar=nil then func^.paramVar:=mathSystem.findVariable('x');

  SetLength(changeVars,length(func^.variables));
  j:=0;
  for i:=0 to high(func^.variables) do begin
    if high(func^.variables[i]^.renderInfo.values)>=0 then
      mathSystem.setVariableValue(func^.variables[i],func^.variables[i]^.renderInfo.values[0]);
    if length(func^.variables[i]^.renderInfo.values)>1 then begin
      changeVars[j]:=func^.variables[i];
      inc(j);
    end;
  end;

  if j=0 then renderWithVar(func^.term,func^.paramVar)
  else begin
    SetLength(selectedValues,j);
    while true do begin
      r:=0;g:=0;b:=0;
      if assigned(outputFunctions.setcolor) then begin
        for i:=0 to j-1 do begin
          fac:=selectedValues[i]/high(changeVars[i]^.renderInfo.values);
          r:=r+fac*GetRValue(changeVars[i]^.renderInfo.maxColor)+(1-fac)*GetRValue(changeVars[i]^.renderInfo.minColor);
          g:=g+fac*GetGValue(changeVars[i]^.renderInfo.maxColor)+(1-fac)*GetGValue(changeVars[i]^.renderInfo.minColor);
          b:=b+fac*GetBValue(changeVars[i]^.renderInfo.maxColor)+(1-fac)*GetBValue(changeVars[i]^.renderInfo.minColor);
        end;
        if r<0 then r:=0;if r>255 then r:=255;
        if g<0 then g:=0;if g>255 then g:=255;
        if b<0 then b:=0;if b>255 then b:=255;
        br:=trunc(r);
        bg:=trunc(g);
        bb:=trunc(b);
        currentColor:=rgb(br,bg,bb);
        outputFunctions.setcolor(currentColor);
      end;
      renderWithVar(func^.term,func^.paramVar);

      inc(selectedValues[0]);
      i:=0;
      while selectedValues[i]>high(changeVars[i]^.renderInfo.values) do begin
        selectedValues[i]:=0;
        mathSystem.setVariableValue(changeVars[i],changeVars[i]^.renderInfo.values[selectedValues[i]]);
        inc(i);
        if i>high(selectedValues) then exit;
        inc(selectedValues[i]);
      end;
      mathSystem.setVariableValue(changeVars[i],changeVars[i]^.renderInfo.values[selectedValues[i]]);
    end;
  end;
end;
procedure TVirtualRenderer.renderWithVar(term: pterm; changeVar: PVariable);
var x,y,oldy: longint;
    yd,oldyd,old2yd,resolution:double;
    hasBeenVisible: boolean;
    defined:boolean;
begin
  with outputOptions, outputFunctions do
    if  changeVar=nil then begin
      yd:=mathSystem.approximate(term)*factorY+null.y;
      moveto(size.left,round(yd));
      lineto(size.right,round(yd));
    end else begin
{      outputFunctions.moveTo(size.Left,size.top);;
      outputFunctions.lineTo(size.Right,size.top);

      lineTo(size.Right,size.Bottom);                  lineTo(size.Right,size.Bottom);
      lineTo(size.Left,size.Bottom);
      lineTo(size.Left,size.top);}

      resolution:=0.001/abs(factorY);

      hasBeenVisible:=false;
      oldy:=-1;
      old2yd:=nan;
      yd:=nan;
      oldyd:=nan;
      defined:=true;
      for x:=size.Left+1 to size.Right do begin
        try
          mathSystem.setVariableValue(changeVar,(x-null.x)/factorX);
          if showExtrema then begin
            old2yd:=oldyd;
            oldyd:=yd;
          end;
          yd:=mathSystem.approximate(term)*factorY+null.y;
          if showExtrema and (oldy<>-1) and (defined){and(old2yd<>nan)}  then begin
            if (size.Left+2<x) {and hasBeenVisible and oldHasBeenVisible }then
              if ((old2yd<oldyd)<>(oldyd<yd)) and
                 ((abs(old2yd-oldyd)>resolution)or(abs(oldyd-yd)>resolution))
                 and((((old2yd<0)=not(yd>size.bottom))
                   and((old2yd>size.bottom)=not(yd<0)))
                  or ((oldyd>=0)and(oldyd<=size.bottom)and
                      (yd>=0)and(yd<=size.bottom))) then begin

                {and(( (old2yd<0)=not(yd>size.bottom))
                                             or( (old2yd>size.bottom)=not(yd<0))}
                setColor(extremaColor);
                moveTo(x-1,null.y);
                lineTo(x-1,oldy);
                setColor(currentColor);
              end;
          end;
          if (yd<=size.Bottom) and (yd>=size.Top) then begin
            y:=round(yd);
            if hasBeenVisible and defined then lineTo(x,y)
            else if defined then begin
              if oldy<>-1 then begin
                if oldy>=size.Bottom then moveTo(x-1,size.Bottom);
                if oldy<=size.top then moveTo(x-1,size.top);
                lineto(x,y);
              end else moveto(x,y);
            end else moveto(x,y);
            hasBeenVisible:=true;
            oldy:=y;
          end else begin
            if hasBeenVisible and defined then begin
              if yd>size.Bottom then lineTo(x,size.Bottom);
              if yd<size.top then lineTo(x,size.top);
            end;
            hasBeenVisible:=false;
            if yd>size.Bottom then oldy:=size.bottom+10;
            if yd<size.top then oldy:=size.top-10;
          end;
        {  end else begin
            y:=round(yd);
            draw(x,y);
            draw:=moveto;
          end;     }
          defined:=true;
        except
          defined:=false;
        end;
      end;
    end;
end;

function TVirtualRenderer.pixelPosToCoordPos(const x,y:longint):tpoint64f;
begin
  with outputOptions do begin
    result.x:=(x-null.x)/factorX;
    result.y:=(y-null.y)/factorY;
  end;
end;
function TVirtualRenderer.pixelPosToCoordPos(const p:tpoint):tpoint64f;
begin
  result:=pixelPosToCoordPos(p.x,p.y);
end;
function TVirtualRenderer.coordPosToPixelPos(const x,y:double):tpoint;
begin
  with outputOptions do begin
    result.x:=round(x*factorX+null.x);
    result.y:=round(y*factorY+null.y);
  end;
end;
function TVirtualRenderer.coordPosToPixelPos(const p:tpoint64f):tpoint;
begin
  result:=coordPosToPixelPos(p.x,p.y);
end;

procedure TVirtualRenderer.centerTo(const x,y:double);
var p: tpoint;
begin
  p:=coordPosToPixelPos(x,y);
  with outputOptions do begin
    inc(null.x,(size.Right+size.left) div 2-p.x);
    inc(null.y,(size.top+size.Bottom) div 2-p.y);
  end;
end;
procedure TVirtualRenderer.centerTo(const p:TPoint64F);
begin
  centerTo(p.x,p.y);
end;
procedure TVirtualRenderer.zoomTo(const x,y:double; const f: double);
begin
  with outputOptions do begin
    factorX:=factorX*f;
    factorY:=factorY*f;
  end;
  centerTo(x,y);
end;
procedure TVirtualRenderer.zoomTo(const p:TPoint64F; const f: double);
begin
  zoomTo(p.x,p.y,f);
end;




destructor TVirtualRenderer.destroy;
begin
  inherited;
end;
end.
